<?php

namespace Drupal\entity_template_ui\Controller;

use Drupal\Component\Utility\SortArray;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\entity_template\Plugin\EntityTemplate\Builder\BuilderInterface;
use Drupal\entity_template\TemplateBuilderManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

class BuilderController extends ControllerBase {

  /**
   * @var \Drupal\entity_template\TemplateBuilderManager
   */
  protected $builderManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('plugin.manager.entity_template.builder'),
      $container->get('entity_type.manager')
    );
  }

  public function __construct(TemplateBuilderManager $template_builder_manager, EntityTypeManagerInterface $entity_type_manager) {
    $this->entityTypeManager = $entity_type_manager;
    $this->builderManager = $template_builder_manager;
  }

  /**
   * Get the collection page.
   *
   * @return array
   */
  public function collection() {
    $build = [];

    $build['table'] = [
      '#type' => 'table',
      '#header' => [
        $this->t("Builder"),
        $this->t("Builds"),
        $this->t("Requires"),
        $this->t("Operations"),
      ],
      '#empty' => $this->t("No builders available."),
    ];

    foreach ($this->builderManager->getDefinitions() as $id => $definition) {
      /** @var \Drupal\entity_template\Plugin\EntityTemplate\Builder\BuilderInterface $plugin */
      $plugin = $this->builderManager->createInstance($id);

      $row = [
        'builder' => [
          '#markup' => $definition['label'],
        ],
        'builds' => [
          '#markup' => $plugin->getReturnTypeLabel(),
        ],
        'requires' => $this->buildParameters($plugin),
        'operations' => $this->buildOperations($plugin),
      ];

      $build['table'][$id] = $row;
    }

    return $build;
  }

  /**
   * Build parameters
   *
   * @param \Drupal\entity_template\Plugin\EntityTemplate\Builder\BuilderInterface $builder
   *
   * @return array
   */
  protected function buildParameters(BuilderInterface $builder) {
    /** @var \Drupal\entity_template\Entity\TemplateBuilder $entity */
    $build = [
      '#theme' => 'item_list',
      '#list_type' => 'ul',
      '#items' => [],
    ];

    foreach ($builder->getContextDefinitions() as $name => $definition) {
      $data_definition = $definition->getDataDefinition();
      $build['#items'][] = $this->t(
        '@label (@type)',
        [
          '@label' => $definition->getLabel(),
          '@type' => $data_definition['label'],
        ]
      );
    }

    return $build;
  }

  /**
   * Build the operations for a given row.
   *
   * @param \Drupal\entity_template\Plugin\EntityTemplate\Builder\BuilderInterface $builder
   *
   * @return array
   */
  protected function buildOperations(BuilderInterface $builder) {
    $operations = $builder->getAdminOperations();
    uasort($operations, SortArray::class.'::sortByWeightElement');

    return [
      '#type' => 'operations',
      '#links' => $operations,
    ];
  }

}
