<?php

namespace Drupal\entity_template_ui\Controller;

use Drupal\Core\Ajax\AjaxHelperTrait;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Plugin\Context\ContextHandlerInterface;
use Drupal\Core\Url;
use Drupal\entity_template\Plugin\EntityTemplate\Template\TemplateInterface;
use Drupal\entity_template\BlueprintStorageInterface;
use Drupal\entity_template\TemplateComponentManager;
use Drupal\entity_template\TemplateManager;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

class ChooseComponentController extends ControllerBase {
  use AjaxHelperTrait;

  /**
   * @var \Drupal\entity_template\TemplateManager
   */
  protected $templateManager;

  /**
   * @var \Drupal\entity_template\TemplateComponentManager
   */
  protected $componentManager;

  /**
   * @var \Drupal\Core\Plugin\Context\ContextHandlerInterface
   */
  protected $contextHandler;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('plugin.manager.entity_template.template'),
      $container->get('plugin.manager.entity_template.component'),
      $container->get('context.handler')
    );
  }

  /**
   * ChooseTemplateController constructor.
   *
   * @param \Drupal\entity_template\TemplateManager $template_manager
   */
  public function __construct(TemplateManager $template_manager, TemplateComponentManager $component_manager, ContextHandlerInterface $context_handler) {
    $this->templateManager = $template_manager;
    $this->componentManager = $component_manager;
    $this->contextHandler = $context_handler;
  }

  /**
   * @param \Drupal\entity_template\BlueprintStorageInterface $blueprint_storage
   */
  public function build(BlueprintStorageInterface $blueprint_storage, $template_id, $sub_template = NULL) {
    $template = $blueprint_storage->getTemplate($template_id);
    if ($sub_template) {
      $path = explode(':', $sub_template);
      while ($bit = array_shift($path)) {
        $template = $template->getComponent($bit);

        if (!$template || !($template instanceof TemplateInterface)) {
          throw new NotFoundHttpException();
        }
      }
    }

    $available_definitions = $this->componentManager->getApplicableDefinitions($template->getReturnType());
    $available_definitions = $this->contextHandler->filterPluginDefinitionsByContexts(
      $template->getTemplateContexts(TRUE),
      $available_definitions
    );

    $build['component_categories'] = [
      '#type' => 'container',
    ];

    foreach ($available_definitions as $id => $available_definition) {
      $category = isset($available_definition['category']) ? $available_definition['category'] : 'Other';

      if (!isset($build[$category])) {
        $build[$category]['#type'] = 'details';
        $build[$category]['#open'] = TRUE;
        $build[$category]['#title'] = $category;
        $build[$category]['links'] = [
          '#theme' => 'links',
          '#links' => [],
        ];
      }

      $link_route = 'entity_template_ui.edit_template.add_component';
      $link_parameters = [
        'template_blueprint_provider' => $blueprint_storage->getProvider()->getPluginId(),
        'blueprint_storage' => $blueprint_storage->getProviderKey(),
        'template_id' => $template_id,
        'plugin_id' => $id,
      ];
      if (!empty($sub_template)) {
        $link_route .= '.sub';
        $link_parameters['sub_template'] = $sub_template;
      }

      $attributes = $this->getAjaxAttributes();
      $build[$category]['links']['#links'][] = [
        'title' => $available_definition['label'],
        'url' => Url::fromRoute($link_route, $link_parameters),
        'attributes' => $attributes,
      ];
    }

    return $build;
  }

  /**
   * Get dialog attributes if an ajax request.
   *
   * @return array
   *   The attributes array.
   */
  protected function getAjaxAttributes() {
    if ($this->isAjax()) {
      return [
        'class' => ['use-ajax'],
        'data-dialog-type' => 'dialog',
        'data-dialog-renderer' => 'off_canvas',
      ];
    }
    return [];
  }
}
