<?php

namespace Drupal\entity_template_ui\Element;

use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Plugin\Context\ContextInterface;
use Drupal\Core\Render\Element\RenderElement;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TypedData\ComplexDataDefinitionInterface;
use Drupal\Core\TypedData\DataDefinitionInterface;
use Drupal\Core\TypedData\ListDataDefinitionInterface;

/**
 * Render the available placeholders from a set of contexts.
 *
 * @RenderElement("available_placeholders")
 */
class AvailablePlaceholders extends RenderElement {

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = get_class($this);
    return [
      '#contexts' => [],
      '#include_filters' => TRUE,
      '#process' => [
        // Core's #process for groups, don't remove it.
        [$class, 'processGroup'],
      ],
      '#pre_render' => [
        [$class, 'preRenderBuildPlaceholderSets'],
        // Core's #pre_render for groups, don't remove it.
        [$class, 'preRenderGroup'],
      ],
    ];
  }

  /**
   * Build the placeholder sets in the element.
   *
   * @param array $element
   *   The available placeholders element.
   *
   * @return array
   *   The fully constructed element.
   */
  public static function preRenderBuildPlaceholderSets(array $element) : array {
    $build = [
      '#type' => 'details',
      '#title' => $element['#title'],
    ];

    /** @var ContextInterface $context */
    foreach ($element['#contexts'] as $name => $context) {
      $build[$name] = [
        '#type' => 'details',
        '#title' => new TranslatableMarkup(
          '@context Placeholders',
          ['@context' =>  $context->getContextDefinition()->getLabel()]
        ),
        'table' => [
          '#lazy_builder' => [static::class . '::buildPlaceholdersTable', [$name, serialize($context->getContextDefinition())]],
          '#create_placeholder' => TRUE,
        ],
      ];
    }

    if (!empty($element['#include_filters'])) {
      $build['filters'] = [
        '#type' => 'details',
        '#title' => new TranslatableMarkup('Filters'),
        '#description' => new TranslatableMarkup(
          'Filters can be used after a placeholder to format the value, for example to print the date in a certain form {{ system.date | format_data("Y-m-d") }}',
        ),
        'table' => [
          '#theme' => 'table',
          '#header' => [
            new TranslatableMarkup('Filter'),
            new TranslatableMarkup('Description'),
          ],
          '#rows' => [],
        ],
      ];
      $filter_manager = \Drupal::service('plugin.manager.typed_data_filter');
      foreach ($filter_manager->getDefinitions() as $id => $definition) {
        $build['filters']['table']['#rows'][] = [
          $id,
          $definition['label'],
        ];
      }
    }

    $element['content'] = $build;
    return $element;
  }

  /**
   * Build the placeholders table for a given context definition.
   *
   * @param string $name
   *   The context name.
   * @param string $context_definition
   *   The context definition.
   *
   * @return array
   *   The render array for the placeholders table.
   */
  public static function buildPlaceholdersTable(string $name, string $context_definition) {
    $context_definition = unserialize($context_definition);

    return [
      '#theme' => 'table',
      '#header' => [
        new TranslatableMarkup('Variable'),
        new TranslatableMarkup('Placeholder'),
        new TranslatableMarkup('Type'),
      ],
      '#rows' => static::getPlaceholders(
        $context_definition->getDataDefinition(),
        $name,
        $context_definition->getLabel()
      ),
    ];
  }

  /**
   * Get the placeholders associated with a given data definition.
   *
   * @param \Drupal\Core\TypedData\DataDefinitionInterface $data_definition
   *   The data definition to get placeholders for.
   * @param string $path
   *   The path so far to the data definition.
   * @param string $label
   *   The label so far for the data definition.
   *
   * @return array[]
   *   An array of arrays of placeholder information, keyed by path. The keys of
   *   each item are:
   *   - label : The label for the placeholder.
   *   - path : The path for the placeholder, wrapped in double curly braces.
   *   - type : The data type.
   */
  protected static function getPlaceholders(DataDefinitionInterface $data_definition, string $path, string $label) {
    $placeholders = [
      $path => [
        'label' => $label,
        'path' => "{{ " . $path . " }}",
        'type' => $data_definition->getDataType(),
      ]
    ];

    if (substr_count($path, '.') > 6 && !($data_definition instanceof ListDataDefinitionInterface)) {
      if ($data_definition instanceof ComplexDataDefinitionInterface) {
        $placeholders += [
          $path . '...' => [
            'path' => "{{ " . $path . "... }}",
            'label' => new TranslatableMarkup('@data properties', ['@data' => $label]),
            'type' => '',
          ]
        ];
      }

      return $placeholders;
    }

    $index_key = '%index';
    if ($data_definition instanceof FieldDefinitionInterface) {
      if ($data_definition->getFieldStorageDefinition()->getCardinality() == 1) {
        $data_definition = $data_definition->getItemDefinition();
      }
      else {
        $index_key = '%delta';
      }
    }

    if ($data_definition instanceof ComplexDataDefinitionInterface) {
      foreach ($data_definition->getPropertyDefinitions() as $property => $property_definition) {
        $placeholders += static::getPlaceholders(
          $property_definition,
          "{$path}.{$property}",
          "{$label} > {$property_definition->getLabel()}",
        );
      }
    }
    else if ($data_definition instanceof ListDataDefinitionInterface) {
      $placeholders += static::getPlaceholders(
        $data_definition->getItemDefinition(),
        "{$path}.{$index_key}",
        "{$label} > List Index"
      );
    }

    return $placeholders;
  }

}
