<?php

namespace Drupal\entity_template_ui\Form;

use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeBundleInfo;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TypedData\TypedDataManagerInterface;
use Drupal\entity_template\BlueprintEntityStorageAdaptor;
use Drupal\entity_template\Entity\TemplateBuilder;
use Drupal\entity_template\TemplateBuilderManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

class TemplateBuilderForm extends EntityForm {

  /**
   * @var \Drupal\Core\Entity\EntityTypeBundleInfo
   */
  protected $entityBundleInfo;

  /**
   * @var \Drupal\Core\TypedData\TypedDataManagerInterface
   */
  protected $typedDataManager;

  /**
   * @var \Drupal\entity_template\TemplateBuilderManager
   */
  protected $builderManager;

  /**
   * @var \Drupal\entity_template\Entity\TemplateBuilder
   */
  protected $entity;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info'),
      $container->get('typed_data_manager'),
      $container->get('plugin.manager.entity_template.builder')
    );
  }

  /**
   * TemplateBuilderForm constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   * @param \Drupal\Core\Entity\EntityTypeBundleInfo $entity_type_bundle_info
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    EntityTypeBundleInfo $entity_type_bundle_info,
    TypedDataManagerInterface $typed_data_manager,
    TemplateBuilderManager $builder_manager
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->entityBundleInfo = $entity_type_bundle_info;
    $this->typedDataManager = $typed_data_manager;
    $this->builderManager = $builder_manager;
  }

  /**
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);
    $form['#tree'] = TRUE;

    $form['label'] = [
      '#title' => $this->t('Label'),
      '#type' => 'textfield',
      '#default_value' => $this->entity->label(),
      '#description' => t('The builder name.'),
      '#required' => TRUE,
      '#size' => 30,
    ];
    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $this->entity->id(),
      '#maxlength' => 128,
      '#machine_name' => [
        'exists' => TemplateBuilder::class.'::load',
        'source' => ['label'],
      ],
    ];

    $form['return_type'] = [
      '#title' => $this->t('Return Type'),
      '#type' => 'select',
      '#options' => $this->getReturnTypeOptions(),
      '#default_value' => $this->entity->get('return_type'),
      '#required' => TRUE,
    ];

    $context_type_options = $this->getContextTypeOptions();
    $form['parameters_wrapper'] = [
      '#title' => $this->t('Parameters'),
      '#description' => $this->t('Configure any parameters that must be provided to the builder.'),
      '#type' => 'fieldset',
    ];
    $form['parameters_wrapper']['parameters'] = [
      '#prefix' => '<div id="parameters-table-wrapper">',
      '#suffix' => '</div>',
      '#parents' => ['parameters'],
      '#type' => 'table',
      '#tree' => TRUE,
      '#header' => [
        $this->t('Label'),
        $this->t('Machine-name'),
        $this->t('Type'),
        $this->t('Operations'),
      ],
    ];
    foreach ($this->entity->getParameters() as $machine_name => $info) {
      $row = [];
      $row['label'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Label'),
        '#title_display' => 'invisible',
        '#default_value' => $info['label'],
      ];
      $row['machine_name'] = [
        '#type' => 'machine_name',
        '#title' => $this->t('Machine Name'),
        '#title_display' => 'invisible',
        '#default_value' => $machine_name,
        '#machine_name' => [
          'label' => '',
          'source' => [ 'parameters_wrapper', 'parameters', $machine_name, 'label'],
          'exists' => [ static::class, 'parameterMachineNameExists' ],
          'standalone' => TRUE,
        ],
        '#disabled' => TRUE,
      ];
      $row['type'] = [
        '#type' => 'select',
        '#title' => $this->t('Type'),
        '#title_display' => 'invisible',
        '#options' => $context_type_options,
        '#default_value' => $info['type']
      ];
      $row['operations'] = [
        '#type' => 'container',
        'remove' => [
          '#type' => 'submit',
          '#name' => 'remove_'.$machine_name,
          '#param_machine_name' => $machine_name,
          '#value' => $this->t('Remove'),
          '#limit_validation_errors' => [],
          '#ajax' => [
            'wrapper' => 'parameters-table-wrapper',
            'callback' => [static::class, 'formAjaxReloadParameters'],
          ],
          '#submit' => [
            '::formSubmitRemoveParameter',
          ],
        ],
      ];

      $form['parameters_wrapper']['parameters'][$machine_name] = $row;
    }

    $row = [];
    $row['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Label'),
      '#title_display' => 'invisible',
    ];
    $row['machine_name'] = [
      '#type' => 'machine_name',
      '#title' => $this->t('Machine Name'),
      '#title_display' => 'invisible',
      '#required' => FALSE,
      '#machine_name' => [
        'label' => '',
        'source' => [ 'parameters_wrapper', 'parameters', '_add_new', 'label'],
        'exists' => [ static::class, 'parameterMachineNameExists' ],
        'standalone' => TRUE,
      ],
    ];
    $row['type'] = [
      '#type' => 'select',
      '#title' => $this->t('Type'),
      '#title_display' => 'invisible',
      '#options' => $context_type_options,
    ];
    $row['operations'] = [
      '#type' => 'container',
      'add' => [
        '#type' => 'submit',
        '#value' => $this->t('Add'),
        '#limit_validation_errors' => [
          ['parameters', '_add_new'],
        ],
        '#ajax' => [
          'wrapper' => 'parameters-table-wrapper',
          'callback' => [static::class, 'formAjaxReloadParameters'],
        ],
        '#validate' => [
          '::formValidateAddParameter',
        ],
        '#submit' => [
          '::formSubmitAddParameter',
        ],
      ],
    ];
    $form['parameters_wrapper']['parameters']['_add_new'] = $row;

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  protected function copyFormValuesToEntity(EntityInterface $entity, array $form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    foreach ($values as $key => $value) {
      if ($key == 'parameters' && isset($value['_add_new'])) {
        unset($value['_add_new']);

        foreach ($value as $machine_name => &$param) {
          unset($param['operations']);
        }
      }

      $entity->set($key, $value);
    }
  }

  /**
   * Validate the information entered for the new context.
   *
   * @param $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function formValidateAddParameter($form, FormStateInterface $form_state) {
    $values = $form_state->getValue(['parameters', '_add_new']);
    $row = &$form['parameters_wrapper']['parameters']['_add_new'];
    if (empty($values['machine_name']) || $this->entity->hasParameter($values['machine_name'])) {
      $form_state->setError($row['machine_name'], $this->t('Context requires a unique machine name.'));
    }
    if (empty($values['label'])) {
      $form_state->setError($row['label'], $this->t('Context requires a label.'));
    }
  }

  /**
   * Submit to add a required context.
   *
   * @param $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function formSubmitAddParameter($form, FormStateInterface $form_state) {
    $values = $form_state->getValue(['parameters', '_add_new']);
    $this->entity->addParameter(
      $values['machine_name'],
      [
        'label' => $values['label'],
        'machine_name' => $values['machine_name'],
        'type' => $values['type'],
      ]
    );
    $form_state->setRebuild(TRUE);

    $user_input = &$form_state->getUserInput();
    unset($user_input['parameters']['_add_new']);
  }

  /**
   * Submit to remove a parameter.
   *
   * @param $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function formSubmitRemoveParameter($form, FormStateInterface $form_state) {
    $button = $form_state->getTriggeringElement();
    $this->entity->removeParameter($button['#param_machine_name']);
    $form_state->setRebuild(TRUE);
  }

  /**
   * Ajax callback to reload the parameters.
   *
   * @param $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array
   */
  public static function formAjaxReloadParameters($form, FormStateInterface $form_state) {
    return $form['parameters_wrapper']['parameters'];
  }

  /**
   * Check whether the machine name of a required context exists already.
   *
   * @param $value
   * @param $element
   * @param $form_state
   *
   * @return boolean
   */
  public static function parameterMachineNameExists($value, $element, FormStateInterface $form_state) {
    return FALSE;
  }

  /**
   * An array of available context types.
   *
   * @return string[][]
   */
  protected function getContextTypeOptions() {
    $context_type_options = [];

    $types = $this->typedDataManager->getDefinitions();
    foreach ($types as $type => $definition) {
      $category = $this->t('Data');
      if (!empty($definition['deriver']) && !empty($types[$definition['id']])) {
        $category = $types[$definition['id']]['label'];
      }
      $context_type_options[(string) $category][$type] = $definition['label'];
    }

    return $context_type_options;
  }

  /**
   * An array of return type options.
   *
   * @return string[][]
   */
  protected function getReturnTypeOptions() {
    $return_type_options = [
      'entity' => $this->t('Any Entity'),
      'entity[]' => $this->t('List of Entities'),
    ];
    foreach ($this->entityTypeManager->getDefinitions() as $id => $definition) {
      $label_string = $definition->getLabel()->render();

      $return_type_options[$label_string]["entity:{$id}"] = $definition->getLabel();
      $return_type_options[$label_string]["entity:{$id}[]"] = $this->t('List of @label', ['@label' => $definition->getPluralLabel()]);

      if (is_subclass_of($definition->getClass(), FieldableEntityInterface::class)) {
        if ($definition->hasKey('bundle')) {
          $return_type_options[$label_string]["entity:{$id}"] = $this->t('Any @label', [
            '@label' => $definition->getLabel(),
          ]);

          foreach ($this->entityBundleInfo->getBundleInfo($id) as $bundle => $bundle_def) {
            $return_type_options[$label_string]["entity:{$id}:{$bundle}"] = $bundle_def['label'] . ' ' . $definition->getLabel();
            $return_type_options[$label_string]["entity:{$id}:{$bundle}[]"] = $this->t(
              'List of @bundle @label',
              [
                '@bundle' => $bundle_def['label'],
                '@label' => $definition->getPluralLabel(),
              ]
            );
          }
        }
        else {
          $return_type_options[$label_string]["entity:{$id}:{$id}"] = $return_type_options[$label_string]["entity:{$id}"];
          unset($return_type_options[$label_string]["entity:{$id}"]);
          $return_type_options[$label_string]["entity:{$id}:{$id}[]"] = $return_type_options[$label_string]["entity:{$id}[]"];
          unset($return_type_options[$label_string]["entity:{$id}[]"]);
        }
      }
    }

    return $return_type_options;
  }

  /**
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return int
   */
  public function save(array $form, FormStateInterface $form_state) {
    // If this is a new builder, then redirect to the default blueprint page.
    $redirect_to_blueprint = FALSE;
    if ($this->entity->isNew()) {
      if (!$this->entity->get('default_blueprint')) {
        $this->entity->set(
          'default_blueprint',
          $this->entity->id() . '.default_blueprint'
        );
        $redirect_to_blueprint = TRUE;
      }
    }

    $return = parent::save($form, $form_state);

    // Clear the cached definitions as this builder entity will likely change
    // which builder plugins are available.
    $this->builderManager->clearCachedDefinitions();

    if ($redirect_to_blueprint) {
      $operations = $this->builderManager
        ->createInstance('config:'.$this->entity->id())
        ->getAdminOperations();

      $form_state->setRedirectUrl($operations['edit-default-blueprint']['url']);
      $this->messenger()->addStatus(new TranslatableMarkup(
        'Builder %builder created successfully. Please configure the default blueprint.',
        [
          '%builder' => $this->entity->label(),
        ]
      ));
    }
    else {
      $form_state->setRedirect('entity_template_ui.builder.collection');
      $this->messenger()->addStatus(new TranslatableMarkup(
        'Builder %builder updated successfully.',
        [
          '%builder' => $this->entity->label(),
        ]
      ));
    }

    return $return;
  }

}
