<?php

namespace Drupal\entity_template\Controller;

use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityFormBuilderInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\entity_template\EntityBuilderTempStoreRepository;
use Drupal\entity_template\TemplateBuilderManager;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Session\SessionInterface;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Controller for selecting and displaying templates.
 */
class EntityTemplateController extends ControllerBase {

  /**
   * The builder manager.
   *
   * @var \Drupal\entity_template\TemplateBuilderManager
   */
  protected $builderManager;

  /**
   * The session.
   *
   * @var \Symfony\Component\HttpFoundation\Session\Session
   */
  protected $session;

  /**
   * The UUID service.
   *
   * @var \Drupal\Component\Uuid\UuidInterface
   */
  protected $uuidService;

  /**
   * The entity builder tempstore.
   *
   * @var \Drupal\entity_template\EntityBuilderTempStoreRepository
   */
  protected $entityBuilderTempStore;

  /**
   * The entity repository.
   *
   * @var \Drupal\Core\Entity\EntityRepositoryInterface
   */
  protected $entityRepository;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity.form_builder'),
      $container->get('current_user'),
      $container->get('plugin.manager.entity_template.builder'),
      $container->get('session'),
      $container->get('uuid'),
      $container->get('entity_template.entity_builder_tempstore'),
      $container->get('entity.repository')
    );
  }

  /**
   * TemplateSelectController constructor.
   *
   * @param \Drupal\Core\Entity\EntityFormBuilderInterface $entity_form_builder
   *   The entity form builder service.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   * @param \Drupal\entity_template\TemplateBuilderManager $template_builder_manager
   *   The template builder manager.
   * @param \Symfony\Component\HttpFoundation\Session\SessionInterface $session
   *   The session.
   * @param \Drupal\Component\Uuid\UuidInterface $uuid_service
   *   The UUID service.
   * @param \Drupal\entity_template\EntityBuilderTempStoreRepository $entity_builder_tempstore
   *   The entity builder tempstore.
   * @param \Drupal\Core\Entity\EntityRepositoryInterface $entity_repository
   *   The entity repository.
   */
  public function __construct(
    EntityFormBuilderInterface $entity_form_builder,
    AccountInterface $current_user,
    TemplateBuilderManager $template_builder_manager,
    SessionInterface $session,
    UuidInterface $uuid_service,
    EntityBuilderTempStoreRepository $entity_builder_tempstore,
    EntityRepositoryInterface $entity_repository
  ) {
    $this->entityFormBuilder = $entity_form_builder;
    $this->currentUser = $current_user;
    $this->builderManager = $template_builder_manager;
    $this->session = $session;
    $this->uuidService = $uuid_service;
    $this->entityBuilderTempStore = $entity_builder_tempstore;
    $this->entityRepository = $entity_repository;
  }

  /**
   * Shows a list of builders.
   *
   * @return array
   *   The built list of builders.
   */
  public function selectBuilder() {
    $build = [
      '#theme' => 'item_list',
      '#attributes' => [
        'class' => ['card-list'],
      ],
      '#items' => [],
    ];

    foreach ($this->builderManager->getDefinitions() as $id => $definition) {
      $params = ['builder_name' => $id];

      $build['#items'][] = [
        'title' => [
          '#type' => 'html_tag',
          '#tag' => 'h3',
          '#value' => $definition['label'],
        ],
        'link' => [
          '#type' => 'link',
          '#url' => Url::fromRoute($this->getBuilderParameterSelectionRoute(), $params),
          '#title' => new TranslatableMarkup('Select'),
          '#attributes' => [
            'class' => ['button', 'button--small'],
          ],
        ],
      ];
    }

    return $build;
  }

  /**
   * Shows a list of template blueprints.
   *
   * @param string $builder_name
   *   The builder name.
   * @param string $session_key
   *   The session key.
   *
   * @return array|\Symfony\Component\HttpFoundation\RedirectResponse
   *   The built list of template blueprints.
   */
  public function selectTemplateBlueprint(string $builder_name, string $session_key = NULL) {
    if (!$session_key) {
      $session_key = $this->session->get('entity_builder_' . $builder_name . '_session_key');
    }

    $build = [
      '#theme' => 'item_list',
      '#attributes' => [
        'class' => ['card-list'],
      ],
      '#items' => [],
    ];

    $builder = $this->builderManager->createInstance($builder_name);

    $params = [
      'builder_name' => $builder_name,
      'session_key' => $session_key,
    ];

    // If there is only one blueprint available, skip template blueprint select.
    if (count($builder->getAvailableBlueprints()) === 1) {
      $params['blueprint_key'] = array_key_first($builder->getAvailableBlueprints());
      return new RedirectResponse(Url::fromRoute($this->getEntityTempStoreRoute(), $params)->toString());
    } else {
      foreach ($builder->getAvailableBlueprints() as $blueprint_key => $blueprint) {
        $params['blueprint_key'] = $blueprint_key;

        $build['#items'][] = [
          'title' => [
            '#type' => 'html_tag',
            '#tag' => 'h3',
            '#value' => $blueprint->label(),
          ],
          'link' => [
            '#type' => 'link',
            '#url' => Url::fromRoute($this->getEntityTempStoreRoute(), $params),
            '#title' => new TranslatableMarkup('Select'),
            '#attributes' => [
              'class' => ['button', 'button--small'],
            ],
          ],
        ];
      }
      return $build;
    }
  }

  /**
   * Gets the builder parameter selection route.
   *
   * @return string
   *   The route.
   */
  protected function getBuilderParameterSelectionRoute() {
    return 'entity_template.select_builder_parameters';
  }

  /**
   * Gets the entity tempstore route.
   *
   * @return string
   *   The route.
   */
  protected function getEntityTempStoreRoute() {
    return 'entity_template.edit.entity_tempstore';
  }

  /**
   * Display entity form.
   *
   * @param string $builder_name
   *   The builder name.
   * @param string $session_key
   *   The session key.
   * @param string $entity_type
   *   The entity type.
   * @param string $entity_uuid
   *   The entity uuid.
   *
   * @return array
   *   The processed form for the given entity.
   *
   * @throws \Exception
   */
  public function displayEntityForm(string $builder_name, string $session_key = NULL, string $blueprint_key = NULL) {
    // If session key is not provided by the route,
    // get it from the current user session.
    if (!$session_key) {
      $session_key = $this->session->get('entity_builder_' . $builder_name . '_session_key');
    }

    $builder = $this->builderManager->createInstance($builder_name);
    $parameters = $this->entityBuilderTempStore->getTempstore()->get('builder_parameters__' . $session_key);
    $result = $builder->execute($parameters, \Drupal::currentUser(), $blueprint_key);

    if (empty($result->getItems())) {
      $this->messenger()->addMessage('No items to edit. Make sure template is configured correctly.', 'error');
      return [];
    }

    // @todo CKAPPS-3695 handle case where there is more than one entity.
    // $result_item_key is always empty until CKAPPS-3695 is complete and we
    // have supported multiple entity builders.
    // Default to the first item in the result.
    $result_items = $result->getItems();
    $entity = !empty($result_item_key) ? $result_items[$result_item_key] : reset($result_items);

    // If the result item key did not match a generated entity then we can't do anything so throw a 404.
    if (!$entity) {
      throw new NotFoundHttpException();
    }

    // The entity may already have been saved, so read the saved entity from the database if that is the case.
    if ($loaded = $this->entityRepository->loadEntityByUuid($entity->getEntityTypeId(), $entity->uuid())) {
      $entity = $loaded;
    }

    return $this->entityFormBuilder->getForm($entity, 'default', ['entity_builder_tempstore_delete' => ['builder_parameter__' . $session_key]]);
  }

}
