<?php

namespace Drupal\entity_template\Plugin\EntityTemplate\Builder;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\entity_template\BlueprintEntityAdaptor;
use Drupal\entity_template\BlueprintEntityStorageAdaptor;
use Drupal\entity_template\Entity\TemplateBlueprint;
use Drupal\entity_template\Plugin\EntityTemplate\BlueprintProvider\BlueprintProviderInterface;

/**
 * Trait for builder plugins that have a configurable default blueprint.
 *
 * @package Drupal\entity_template\Plugin\EntityTemplate\Builder
 */
trait ConfigurableDefaultBlueprintTrait {

  /**
   * Get the entity type manager.
   *
   * @return \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected function entityTypeManager() : EntityTypeManagerInterface {
    return $this->entityTypeManager ?: \Drupal::entityTypeManager();
  }

  /**
   * Get the plugin id.
   *
   * @return string
   */
  abstract public function getPluginId();

  /**
   * Get the default blueprint config.
   *
   * @return \Drupal\entity_template\Entity\TemplateBlueprint
   *   The default blueprint entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getDefaultBlueprintConfig() : TemplateBlueprint {
    $blueprint_storage = $this->entityTypeManager()->getStorage('entity_template_blueprint');
    $default_blueprint_id = $this->getPluginId().'.default';

    while (true) {
      /** @var \Drupal\entity_template\Entity\TemplateBlueprint $default_blueprint */
      if ($default_blueprint = $blueprint_storage->load($default_blueprint_id)) {
        // If this blueprint already exists and isn't part of this builder, then
        // try a different name.
        if ($default_blueprint->getBuilder() !== $this->getPluginId()) {
          $default_blueprint_id .= '_';
          continue;
        }

        $default_blueprint->set('is_default', TRUE);
        return $default_blueprint;
      }
      else {
        return $blueprint_storage->create([
          'id' => $default_blueprint_id,
          'label' => $this->t('@builder Default', ['@builder' => $this->getPluginDefinition()['label']]),
          'builder' => $this->getPluginId(),
          'is_default' => TRUE,
        ]);
      }
    }
  }

  /**
   * Get the default blueprint for this builder.
   *
   * @return \Drupal\entity_template\BlueprintInterface
   *   The default blueprint.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getDefaultBlueprint() {
    return new BlueprintEntityAdaptor($this->getDefaultBlueprintConfig());
  }

  /**
   * Get the default blueprint storage.
   *
   * @param \Drupal\entity_template\Plugin\EntityTemplate\BlueprintProvider\BlueprintProviderInterface $provider
   *   The blueprint provider.
   *
   * @return \Drupal\entity_template\BlueprintStorageInterface
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getDefaultBlueprintStorage(BlueprintProviderInterface $provider) {
    return new BlueprintEntityStorageAdaptor($this->getDefaultBlueprintConfig(), $provider);
  }

  /**
   * Get the admin operations fo ConfigurableDefaultBlueprint builders.
   *
   * @return array
   *   The operations.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getAdminOperations() {
    $operations = [];

    if ($storage = $this->getDefaultBlueprintStorage($this->blueprintProviderManager->createInstance("builder"))) {
      $operations['edit-default-blueprint'] = [
        'title' => new TranslatableMarkup('Default Blueprint'),
        'weight' => 50,
        'url' => $storage->getEditUrl(),
      ];
    }

    return $operations;
  }

}
