<?php

namespace Drupal\entity_template\Plugin\EntityTemplate\Template;

use Drupal\Component\Plugin\Context\ContextInterface as ComponentContextInterface;
use Drupal\Component\Plugin\Exception\ContextException;
use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Condition\ConditionManager;
use Drupal\Core\Entity\EntityTypeBundleInfo;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Plugin\Context\Context;
use Drupal\Core\Plugin\Context\ContextDefinition;
use Drupal\Core\Plugin\Context\ContextHandlerInterface;
use Drupal\Core\Plugin\ContextAwarePluginBase;
use Drupal\Core\Plugin\ContextAwarePluginInterface;
use Drupal\Core\Plugin\DefaultLazyPluginCollection;
use Drupal\Core\Plugin\PluginWithFormsInterface;
use Drupal\Core\Plugin\PluginWithFormsTrait;
use Drupal\entity_template\Exception\TemplateException;
use Drupal\entity_template\BlueprintResult;
use Drupal\entity_template\SampleContextValueTrait;
use Drupal\entity_template\TemplateComponentManager;
use Drupal\entity_template\TemplateResult;
use Drupal\entity_template\TypeLabelTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;


abstract class Template extends ContextAwarePluginBase implements TemplateInterface, ContainerFactoryPluginInterface, PluginWithFormsInterface {
  use PluginWithFormsTrait;
  use TemplateComponentTrait;
  use SampleContextValueTrait, TypeLabelTrait {
    SampleContextValueTrait::entityTypeManager insteadof TypeLabelTrait;
    SampleContextValueTrait::entityTypeBundleInfo insteadof TypeLabelTrait;
  }

  /**
   * @var \Drupal\Core\Plugin\Context\ContextInterface[]
   */
  protected $extraContexts = [];

  /**
   * @var \Drupal\Component\Plugin\LazyPluginCollection
   */
  protected $componentPluginCollection;

  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * @var \Drupal\Core\Entity\EntityTypeBundleInfo
   */
  protected $entityTypeBundleInfo;

  /**
   * @var \Drupal\Core\Plugin\Context\ContextHandlerInterface
   */
  protected $contextHandler;

  /**
   * @var \Drupal\entity_template\TemplateComponentManager
   */
  protected $componentManager;

  /**
   * @var \Drupal\Component\Uuid\UuidInterface
   */
  protected $uuidGenerator;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info'),
      $container->get('context.handler'),
      $container->get('plugin.manager.entity_template.component'),
      $container->get('uuid')
    );
  }

  /**
   * Template constructor.
   *
   * @param \Drupal\entity_template\BlueprintInterface $blueprint
   * @param array $configuration
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   * @param \Drupal\Core\Plugin\Context\ContextHandlerInterface $context_handler
   * @param \Drupal\Core\Condition\ConditionManager $condition_manager
   * @param \Drupal\entity_template\TemplateComponentManager $component_manager
   */
  public function __construct(
    $configuration,
    $plugin_id,
    $plugin_definition,
    EntityTypeManagerInterface $entity_type_manager,
    EntityTypeBundleInfo $entity_type_bundle_info,
    ContextHandlerInterface $context_handler,
    TemplateComponentManager $component_manager,
    UuidInterface $uuid_generator
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
    $this->contextHandler = $context_handler;
    $this->componentManager = $component_manager;
    $this->uuidGenerator = $uuid_generator;
    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public function getKey() : string {
    return $this->configuration['uuid'];
  }

  /**
   * {@inheritdoc}
   */
  public function getContexts($sample = FALSE) {
    return $this->getExtraContexts($sample) ?: [];
  }

  /**
   * {@inheritdoc}
   */
  public function getContext($name, $sample = FALSE) {
    return $this->getExtraContext($name, $sample);
  }

  /**
   * @return \Drupal\Core\Plugin\Context\ContextInterface[]
   */
  protected function getExtraContexts($sample = FALSE) {
    return $this->extraContexts;
  }

  /**
   * @param $name
   *
   * @return \Drupal\Core\Plugin\Context\ContextInterface
   */
  protected function getExtraContext($name, $sample = FALSE) {
    if (!isset($this->extraContexts[$name]) && ($def = $this->getExtraContextDefinition($name))) {
      if ($def->getDefaultValue()) {
        $this->extraContexts[$name] = new Context($def, $def->getDefaultValue());
      }
      else if ($sample) {
        $this->extraContexts[$name] = new Context($def, $this->generateSampleValue($def));
      }
    }

    if (!isset($this->extraContexts[$name])) {
      throw new \InvalidArgumentException("No context with name {$name}.");
    }

    return $this->extraContexts[$name];
  }

  /**
   * {@inheritdoc}
   */
  public function setContext($name, ComponentContextInterface $context) {
    if ($this->getExtraContextDefinition($name)) {
      $this->extraContexts[$name] = $context;
    }

    return $this;
  }

  /**
   * Get all the context values.
   *
   * @return array
   */
  public function getContextValues() {
    $values = [];

    foreach ($this->getContextDefinitions() as $name => $definition) {
      $values[$name] = $this->getContext($name)->getContextValue();
    }

    return $values;
  }

  /**
   * Get a particular context value.
   *
   * @param $name
   *
   * @return mixed
   */
  public function getContextValue($name) {
    return $this->getContext($name)->getContextValue();
  }

  /**
 * Get all the context definition interfaces
 *
 * @return \Drupal\Core\Plugin\Context\ContextDefinitionInterface[]
 */
  public function getContextDefinitions() {
    return $this->getExtraContextDefinitions();
  }

  /**
   * Get extra context definitions.
   *
   * @return \Drupal\Core\Plugin\Context\ContextDefinitionInterface[]
   */
  protected function getExtraContextDefinitions() {
    return [];
  }

  /**
   * Get a specific context definition.
   *
   * @param $name
   *
   * @return \Drupal\Core\Plugin\Context\ContextDefinitionInterface
   */
  public function getContextDefinition($name) {
    $definitions = $this->getContextDefinitions();

    if (!isset($definitions[$name])) {
      throw new \InvalidArgumentException("No context definition found for ".$name);
    }

    return $definitions[$name];
  }

  /**
   * Get extra context definition.
   *
   * @param $name
   *
   * @return \Drupal\Core\Plugin\Context\ContextDefinitionInterface
   */
  protected function getExtraContextDefinition($name) {
    $definitions = $this->getExtraContextDefinitions();

    if (!isset($definitions[$name])) {
      throw new \InvalidArgumentException("No context definition found for ".$name);
    }

    return $definitions[$name];
  }

  /**
   * {@inheritdoc}
   */
  public function applies(CacheableMetadata $cache_metadata = NULL) {
    return TRUE;
  }

  /**
   * Execute the template.
   *
   * @return \Drupal\entity_template\TemplateResult
   */
  public function execute() {
    $this->prepareContexts();

    $template_result = new TemplateResult();
    try {
      $entity = $this->createEntity();

      /** @var \Drupal\entity_template\Plugin\EntityTemplate\Component\ComponentInterface $plugin */
      foreach ($this->getComponents() as $plugin) {
        if ($plugin instanceof ContextAwarePluginInterface) {
          $this->contextHandler->applyContextMapping($plugin, $this->getContexts());
        }

        $plugin->apply($entity, $template_result);
      }

      $template_result->setEntity($entity);
    }
    catch (\Exception $exception) {
      $template_result->addMessage($exception->getMessage());
    }

    return $template_result;
  }

  /**
   * Prepare contexts.
   */
  protected function prepareContexts() {
  }

  /**
   * Create the entity initially.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *
   * @throws \Drupal\entity_template\Exception\TemplateException;
   */
  protected function createEntity() {
    $target_type_id = $this->getTargetEntityTypeId();

    $target_type = $this->entityTypeManager->getDefinition($target_type_id);
    $storage = $this->entityTypeManager->getStorage($target_type_id);

    $initial_values = [];
    if ($target_type->hasKey('bundle')) {
      $initial_values[$target_type->getKey('bundle')] = $this->getTargetEntityBundle();
    }

    // Allow the blueprint or builder to massage the initial values.
    if ($this instanceof BlueprintTemplateInterface) {
      $blueprint = $this->getBlueprint();
      if (is_callable([$blueprint, 'massageInitialValues'])) {
        $blueprint->massageInitialValues($initial_values, $target_type_id);
      }

      $builder = $blueprint->getBuilder();
      if (is_callable([$builder, 'massageInitialValues'])) {
        $builder->massageInitialValues($initial_values, $target_type_id, $blueprint);
      }
    }
    return $storage->create(array_filter($initial_values));
  }

  /**
   * Get the target entity type id.
   *
   * @throws \Drupal\entity_template\Exception\TemplateException
   */
  abstract protected function getTargetEntityTypeId();

  /**
   * Get the target entity bundle where applicable
   *
   * @return string
   *
   * @throws \Drupal\entity_template\Exception\TemplateException
   */
  abstract protected function getTargetEntityBundle();

  /**
   * {@inheritdoc}
   */
  public function isSingle() {
    return !empty($this->pluginDefinition['single']);
  }

  /**
   * {@inheritdoc}
   */
  public function getReturnType() {
    return 'entity:'.$this->getTargetEntityTypeId().(($bundle = $this->getTargetEntityBundle()) ? ':'.$bundle : '');
  }

  /**
   * {@inheritdoc}
   */
  public function getReturnTypeLabel() {
    return $this->getTypeLabel($this->getReturnType());
  }

  /**
   * Gets this plugin's configuration.
   *
   * @return array
   *   An array of this plugin's configuration.
   */
  public function getConfiguration() {
    return [
      'id' => $this->getPluginId(),
    ] + $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = [
      'id' => $this->getPluginId(),
    ] + $configuration + $this->defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function label() {
    return !empty($this->configuration['label']) ? $this->configuration['label'] : ' - Untitled -';
  }

  /**
   * Gets default configuration for this plugin.
   *
   * @return array
   *   An associative array with the default configuration.
   */
  public function defaultConfiguration() {
    return [
      'label' => '',
      'description' => '',
      'priority' => 0,
      'conditions' => [],
      'components' => [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getTemplateContexts($sample = FALSE) {
    return $this->getContexts($sample);
  }

  /**
   * {@inheritdoc}
   */
  public function getContextProvidingTemplate() {
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getRootTemplate() {
    return $this;
  }


}
